import unittest

from daktari.result_printer import progress_bar


class TestResultPrinter(unittest.TestCase):
    def test_progress_bar(self):
        self.assertEqual(progress_bar(0, 50, False), "Progress: [>                        ] 0%  (0/50)")
        self.assertEqual(progress_bar(5, 50, False), "Progress: [->                       ] 10%  (5/50)")
        self.assertEqual(progress_bar(10, 40, False), "Progress: [----->                   ] 25%  (10/40)")
        self.assertEqual(progress_bar(20, 40, False), "Progress: [----------->             ] 50%  (20/40)")
        self.assertEqual(progress_bar(20, 37, False), "Progress: [------------>            ] 54%  (20/37)")
        self.assertEqual(progress_bar(30, 40, False), "Progress: [----------------->       ] 75%  (30/40)")
        self.assertEqual(progress_bar(40, 40, False), "Progress: [------------------------>] 100%  (40/40)")
        self.assertEqual(progress_bar(20, 40, True), "Progress: [-----------x             ] 50%  (20/40)")



import re

from daktari.command_utils import get_stdout

from daktari.os import OS
from daktari.check import Check, CheckResult
from daktari.version_utils import get_simple_cli_version
from daktari.file_utils import file_contains_text
from typing import Optional
from os import getcwd


class DirenvInstalled(Check):
    name = "direnv.installed"

    def __init__(self, required_version: Optional[str] = None, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version
        self.suggestions = {
            OS.GENERIC: "Install direnv: https://direnv.net/#getting-started",
        }

    def check(self) -> CheckResult:
        installed_version = get_simple_cli_version("direnv")
        return self.validate_semver_expression(
            "direnv", installed_version, self.required_version, self.recommended_version
        )


class EnvrcContainsText(Check):
    name = "direnv.envrc.containsText"

    def __init__(self, expected_string: str, suggestion: str):
        self.file_path = f"{getcwd()}/.envrc"
        self.expected_string = expected_string
        self.pass_fail_message = f"{self.file_path} does <not/> contain '{expected_string}'"
        self.suggestions = {OS.GENERIC: suggestion}

    def check(self) -> CheckResult:
        return self.verify(file_contains_text(self.file_path, self.expected_string), self.pass_fail_message)


class DirenvAllowed(Check):
    name = "direnv.allowed"
    depends_on = [DirenvInstalled, EnvrcContainsText]

    def __init__(self):
        self.suggestions = {OS.GENERIC: "<cmd>direnv allow .</cmd>"}

    def check(self) -> CheckResult:
        direnv_status = get_stdout("direnv status")
        if direnv_status is None:
            return self.failed("direnv status returned no output")
        cwd = getcwd()
        query = f"Found RC path {cwd}/.envrc\n.*\n.*\nFound RC allowed true"
        direnv_allowed = re.search(query, direnv_status) is not None
        return self.verify(direnv_allowed, f"{cwd} is <not/> allowed to use direnv")



from os.path import expanduser
from typing import List

from daktari.check import Check, CheckResult
from daktari.file_utils import dir_exists, file_exists
from daktari.os import OS


class FilesExist(Check):
    name = "files.exist"
    file_paths: List[str] = []
    pass_fail_message = ""

    def check(self) -> CheckResult:
        files_exist = all([file_exists(expanduser(file_path)) for file_path in self.file_paths])
        return self.verify(files_exist, self.pass_fail_message)


class FileExists(FilesExist):
    name = "file.exists"

    def __init__(self, file_path: str, suggestion: str):
        self.file_paths = [file_path]
        self.pass_fail_message = f"{file_path} is <not/> present"
        self.suggestions = {OS.GENERIC: suggestion}


class DirsExist(Check):
    name = "directories.exist"
    dir_paths: List[str] = []
    pass_fail_message = ""

    def check(self) -> CheckResult:
        dirs_exist = all([dir_exists(expanduser(dir_path)) for dir_path in self.dir_paths])
        return self.verify(dirs_exist, self.pass_fail_message)


class DirExists(DirsExist):
    name = "dir.exists"

    def __init__(self, dir_path: str, suggestion: str):
        self.dir_paths = [dir_path]
        self.pass_fail_message = f"{dir_path} is <not/> present"
        self.suggestions = {OS.GENERIC: suggestion}



import unittest

from daktari.checks.ssh import is_ssh_configured_to_use_macos_keychain


class TestSSH(unittest.TestCase):
    def test_ssh_configured_correctly(self):
        self.assertTrue(is_ssh_configured_to_use_macos_keychain("./daktari/checks/test_ssh_configs/good_ssh_config"))

    def test_ssh_not_configured(self):
        self.assertFalse(is_ssh_configured_to_use_macos_keychain("./daktari/checks/test_ssh_configs/bad_ssh_config"))

    def test_ssh_no_config_file(self):
        self.assertFalse(is_ssh_configured_to_use_macos_keychain("/some/file/path/that/does/not/exist"))

    def test_ssh_half_configured_file(self):
        self.assertFalse(
            is_ssh_configured_to_use_macos_keychain("./daktari/checks/test_ssh_configs/half_configured_ssh_config")
        )

    def test_ssh_incorrectly_configured_file(self):
        self.assertFalse(
            is_ssh_configured_to_use_macos_keychain(
                "./daktari/checks/test_ssh_configs/incorrectly_configured_ssh_config"
            )
        )



import unittest

from semver import VersionInfo

from .java import parse_java_version_output, parse_javac_version_output


class TestJava(unittest.TestCase):
    def test_parse_java_non_semver_format_version(self):
        # JRE versions prior to 9 do not use semver format
        version_output = """openjdk version "1.8.0_292"
OpenJDK Runtime Environment (build 1.8.0_292-8u292-b10-0ubuntu1~20.04-b10)
OpenJDK 64-Bit Server VM (build 25.292-b10, mixed mode)
"""
        version = parse_java_version_output(version_output)
        self.assertEqual(version, VersionInfo(8))

    def test_parse_ubuntu_openjdk_17_version(self):
        version_output = """openjdk version "17" 2021-09-14
OpenJDK Runtime Environment (build 17+35-Ubuntu-121.04)
OpenJDK 64-Bit Server VM (build 17+35-Ubuntu-121.04, mixed mode, sharing)
"""
        version = parse_java_version_output(version_output)
        self.assertEqual(version, VersionInfo(17, 0, 0))

    def test_parse_java_semver_format_version(self):
        version_output = """openjdk version "11.0.9" 2020-10-20
OpenJDK Runtime Environment AdoptOpenJDK (build 11.0.9+11)
OpenJDK 64-Bit Server VM AdoptOpenJDK (build 11.0.9+11, mixed mode)
"""
        version = parse_java_version_output(version_output)
        self.assertEqual(version, VersionInfo(11, 0, 9))

    def test_parse_java_unknown_format_version(self):
        version_output = 'openjdk version "a.b.c"'
        version = parse_java_version_output(version_output)
        self.assertEqual(version, None)

    def test_parse_java_unknown_format_output(self):
        version_output = """

Command 'java' not found, but can be installed with:

sudo apt install openjdk-11-jre-headless  # version 11.0.11+9-0ubuntu2~20.04, or
sudo apt install default-jre              # version 2:1.11-72
sudo apt install openjdk-13-jre-headless  # version 13.0.7+5-0ubuntu1~20.04
sudo apt install openjdk-16-jre-headless  # version 16.0.1+9-1~20.04
sudo apt install openjdk-8-jre-headless   # version 8u292-b10-0ubuntu1~20.04
sudo apt install openjdk-14-jre-headless  # version 14.0.2+12-1~20.04

"""
        version = parse_java_version_output(version_output)
        self.assertEqual(version, None)

    def test_parse_javac_non_semver_format_version(self):
        version_output = "javac 1.8.0_292\n"
        version = parse_javac_version_output(version_output)
        self.assertEqual(version, VersionInfo(8, 0, 0))

    def test_parse_javac_semver_format_version(self):
        version_output = "javac 11.0.9\n"
        version = parse_javac_version_output(version_output)
        self.assertEqual(version, VersionInfo(11, 0, 9))

    def test_parse_javac_unknown_format_version(self):
        version_output = "javac a.b.c\n"
        version = parse_javac_version_output(version_output)
        self.assertEqual(version, None)

    def test_parse_javac_unknown_format_output(self):
        version_output = """
Command 'javac' not found, but can be installed with:

sudo apt install openjdk-11-jdk-headless  # version 11.0.11+9-0ubuntu2~20.04, or
sudo apt install default-jdk              # version 2:1.11-72
sudo apt install openjdk-13-jdk-headless  # version 13.0.7+5-0ubuntu1~20.04
sudo apt install openjdk-16-jdk-headless  # version 16.0.1+9-1~20.04
sudo apt install openjdk-8-jdk-headless   # version 8u292-b10-0ubuntu1~20.04
sudo apt install ecj                      # version 3.16.0-1
sudo apt install openjdk-14-jdk-headless  # version 14.0.2+12-1~20.04

"""
        version = parse_javac_version_output(version_output)
        self.assertEqual(version, None)


if __name__ == "__main__":
    unittest.main()



from daktari.check import Check, CheckResult
from daktari.file_utils import file_contains_text, get_absolute_path
from daktari.os import OS, detect_os


def is_ssh_configured_to_use_macos_keychain(ssh_config_path: str = "~/.ssh/config") -> bool:
    absolute_ssh_config_path = get_absolute_path(ssh_config_path)
    return file_contains_text(absolute_ssh_config_path, "IgnoreUnknown UseKeychain") and file_contains_text(
        absolute_ssh_config_path, "UseKeychain yes"
    )


class SSHConfigSetup(Check):
    name = "ssh.config.setup"

    suggestions = {
        OS.OS_X: """
            Add "IgnoreUnknown UseKeychain" and "UseKeychain yes" to ~/.ssh/config.
            E.g:
                Host *
                    IgnoreUnknown UseKeychain
                    UseKeychain yes
                    IdentityFile ~/.ssh/id_rsa
            """
    }

    def check(self) -> CheckResult:
        if detect_os() != OS.OS_X:
            return self.passed("Setup not required on non-Macbook devices")
        else:
            if not is_ssh_configured_to_use_macos_keychain():
                return self.failed("'IgnoreUnknown UseKeychain' or 'UseKeychain yes' not present in ~/.ssh/config")
            return self.passed("~/.ssh/config setup correctly")



import logging
import os
from datetime import datetime
from OpenSSL import crypto

from daktari.check import Check, CheckResult
from daktari.os import OS


class CertificateIsNotExpired(Check):
    name = "certificate.isNotExpired"

    def __init__(self, certificate_path: str):
        self.certificate_path = certificate_path
        self.suggestions = {
            OS.GENERIC: f"Regenerate the certificate at {certificate_path}",
        }

    def check(self) -> CheckResult:
        with open(self.certificate_path, "rb") as f:
            cert = crypto.load_certificate(crypto.FILETYPE_PEM, f.read())
            logging.debug(f"Raw expiry: {cert.get_notAfter()!r}")

            expiry_bytes = cert.get_notAfter()
            if expiry_bytes is None:
                return self.passed_with_warning(
                    f"Unable to determine expiry date of {os.path.basename(self.certificate_path)}"
                )

            expiry = datetime.strptime(expiry_bytes.decode(), "%Y%m%d%H%M%SZ")
            if expiry > datetime.now():
                return self.passed(f"{os.path.basename(self.certificate_path)} is not expired")
            else:
                return self.failed(f"{os.path.basename(self.certificate_path)} expired on {expiry}")



import logging
import re
from semver import VersionInfo
from typing import Optional

from daktari.check import Check, CheckResult
from daktari.command_utils import get_stdout
from daktari.os import OS
from daktari.version_utils import try_parse_semver


flutter_version_pattern = re.compile(r"Flutter\s+([\d\.]+)")


def parse_flutter_version_output(version_output: Optional[str]) -> Optional[VersionInfo]:
    if version_output:
        match = flutter_version_pattern.search(version_output)
        if match:
            version_string = match.group(1)
            logging.debug(f"Flutter version string: {version_string}")
            return try_parse_semver(version_string)
    return None


def get_flutter_version() -> Optional[VersionInfo]:
    version_output = get_stdout("flutter --version")
    return parse_flutter_version_output(version_output)


class FlutterInstalled(Check):
    name = "flutter.installed"

    suggestions = {
        OS.GENERIC: "Install Flutter: https://flutter.dev/docs/get-started/install",
    }

    def __init__(self, required_version: Optional[str] = None, recommended_version: Optional[str] = None):
        self.required_version = required_version
        self.recommended_version = recommended_version

    def check(self) -> CheckResult:
        flutter_version = get_flutter_version()
        logging.info(f"Flutter version: {flutter_version}")
        return self.validate_semver_expression(
            "Flutter", flutter_version, self.required_version, self.recommended_version
        )



import logging
import re
from dataclasses import dataclass, replace, field
from pathlib import Path
from typing import Any, Dict, List, Optional

import yaml
from colors import red, yellow
from packaging import version
from yaml import YAMLError

from daktari import __version__
from daktari.check import Check
from daktari.check_utils import get_all_dependent_check_names
from daktari.resource_utils import get_resource
from daktari.result_printer import print_suggestion_text


@dataclass
class Config:
    min_version: Optional[str]
    title: Optional[str]
    checks: List[Check]
    ignored_checks: List[Check] = field(default_factory=list)
    quiet_mode: bool = False


version_regex = re.compile('daktari_version.*"([0-9.]+)"')
LOCAL_CONFIG_PATH = ".daktari-local.yaml"
LOCAL_CONFIG_TEMPLATE = "daktari-local-template.yaml"


def read_config(config_path: Path) -> Optional[Config]:
    raw_config = config_path.read_text()
    config = parse_raw_config(config_path, raw_config)
    if config is None:
        return config

    return apply_local_config(config)


def apply_local_config(config: Config) -> Optional[Config]:
    if not Path(LOCAL_CONFIG_PATH).is_file():
        return config

    try:
        with open(LOCAL_CONFIG_PATH, "rb") as local_config_file:
            local_config = yaml.safe_load(local_config_file)
    except YAMLError:
        print(red(f"❌  Failed to parse {LOCAL_CONFIG_PATH} - config is not valid YAML. Error follows."))
        logging.error(f"Exception reading {LOCAL_CONFIG_PATH}", exc_info=True)
        return None

    # E.g. the file has been entirely commented out
    if local_config is None:
        return config

    ignored_checks: List[str] = local_config.get("ignoredChecks", [])
    updated_config = remove_ignored_checks(config, ignored_checks)

    if local_config.get("alwaysQuiet", False):
        updated_config = replace(updated_config, quiet_mode=True)

    return updated_config


def write_local_config_template():
    contents = get_resource(LOCAL_CONFIG_TEMPLATE)
    with open(LOCAL_CONFIG_PATH, "w") as config_file:
        config_file.write(contents)

    print(
        f"A local config file has been generated at {LOCAL_CONFIG_PATH}. "
        f"Use this to override daktari behaviour - see the file for more details."
    )


def remove_ignored_checks(config: Config, ignored_check_names: List[str]) -> Config:
    ignored_checks = [check for check in config.checks if check_should_be_ignored(check, ignored_check_names)]
    remaining_checks = [check for check in config.checks if check not in ignored_checks]
    return replace(config, checks=remaining_checks, ignored_checks=ignored_checks)


def check_should_be_ignored(check: Check, ignored_check_names: List[str]) -> bool:
    dependents = get_all_dependent_check_names(check)
    return check.name in ignored_check_names or any([dependent in ignored_check_names for dependent in dependents])


def parse_raw_config(config_path: Path, raw_config: str) -> Optional[Config]:
    if not check_version_compatibility(config_path, raw_config):
        return None

    variables: Dict[str, Any] = {}
    try:
        compiled_config = compile(raw_config, config_path, "exec")
        exec(compiled_config, variables)
    except Exception:
        print(red(f"❌  Failed to parse {config_path} - config is not valid."))
        logging.error(f"Exception reading {config_path}", exc_info=True)
        return None

    checks = variables.get("checks", [])
    title = variables.get("title", None)
    min_version = variables.get("daktari_version", None)
    return Config(min_version, title, checks)


def check_version_compatibility(config_path: Path, raw_config: str) -> bool:
    match = version_regex.search(raw_config)
    disclaimer = "Specifying daktari_version is recommended to ensure team members have compatible versions installed."
    if match is None:
        print(yellow(f"⚠️  No minimum version found in {config_path}. {disclaimer}"))
        return True

    my_version = version.parse(__version__)
    required_version = version.parse(match.group(1))

    if not isinstance(required_version, version.Version):
        print(red(f"❌  Invalid daktari_version in {config_path}: {required_version}"))
        return False

    logging.debug(f"Doing version check. Mine [{my_version}], required [{required_version}]")
    if required_version > my_version:
        print(
            red(
                f"❌  Installed version of daktari [{my_version}] is "
                f"too old for this project (needs at least {required_version}). "
            )
        )
        print_suggestion_text("pip install daktari -U")
        return False

    return True



import unittest

from daktari.check_sorter import sort_checks
from daktari.check_utils import CyclicCheckException
from daktari.test_check_factory import DummyCheck


class TestCheckSorter(unittest.TestCase):
    def test_raises_error_if_cycle_found(self):
        check_a = DummyCheck("a")
        check_b = DummyCheck("b")

        check_a.depends_on = [check_b]
        check_b.depends_on = [check_a]

        with self.assertRaises(CyclicCheckException):
            sort_checks([check_a, check_b])

    def test_sorts_into_right_order(self):
        parent_check_a = DummyCheck("parent.a")
        sub_check_a_1 = DummyCheck("sub.check.a.1", [parent_check_a])
        sub_check_a_2 = DummyCheck("sub.check.a.2", [parent_check_a])
        sub_sub_check_a = DummyCheck("sub.sub.check.a", [sub_check_a_1, sub_check_a_2])

        parent_check_b = DummyCheck("parent.b")
        sub_check_b_1 = DummyCheck("sub.check.b.1", [parent_check_b])
        sub_check_b_2 = DummyCheck("sub.check.b.2", [parent_check_b])
        sub_check_b_3 = DummyCheck("sub.check.b.3", [parent_check_b])

        sub_check_a_b = DummyCheck("sub.check.a.b", [parent_check_b, parent_check_a])

        free_check_1 = DummyCheck("free.check.1")
        free_check_2 = DummyCheck("free.check.2")

        unsorted_list = [
            sub_check_b_1,
            sub_sub_check_a,
            parent_check_b,
            sub_check_b_3,
            sub_check_a_b,
            free_check_1,
            sub_check_b_2,
            sub_check_a_2,
            parent_check_a,
            free_check_2,
            sub_check_a_1,
        ]

        sorted_checks = [check.name for check in sort_checks(unsorted_list)]
        self.assertGreater(sorted_checks.index("sub.check.a.1"), sorted_checks.index("parent.a"))
        self.assertGreater(sorted_checks.index("sub.check.a.2"), sorted_checks.index("parent.a"))
        self.assertGreater(sorted_checks.index("sub.sub.check.a"), sorted_checks.index("sub.check.a.1"))
        self.assertGreater(sorted_checks.index("sub.sub.check.a"), sorted_checks.index("sub.check.a.2"))

        self.assertGreater(sorted_checks.index("sub.check.b.1"), sorted_checks.index("parent.b"))
        self.assertGreater(sorted_checks.index("sub.check.b.2"), sorted_checks.index("parent.b"))
        self.assertGreater(sorted_checks.index("sub.check.b.3"), sorted_checks.index("parent.b"))

        self.assertGreater(sorted_checks.index("sub.check.a.b"), sorted_checks.index("parent.a"))
        self.assertGreater(sorted_checks.index("sub.check.a.b"), sorted_checks.index("parent.b"))


if __name__ == "__main__":
    unittest.main()
